package dags;

public class DagGrapher{
/** *****************************************************************
 *  Produces a string with which to draw the dag in LaTeX. Requires that
 *  the arrays in Dager be left as they were after construction. So DO
 *  NOT call Dager.reset() until you have finished using
 *  DagGrapher.
 ******************************************************************** */

// ----------------------------------------------------------------------------
// Data
// ----------------------------------------------------------------------------
  
  int[][] dag;
  String[] dfds;
  String[] arities;
  int[][] parentsMulti;
  String[] points;

// ----------------------------------------------------------------------------
// Constructors
// ----------------------------------------------------------------------------

  public DagGrapher(int[][] a) {
    dag = a;
    dfds = Dager.dfds;
    arities = Dager.arities;
    parentsMulti = Dager.parentsMulti;
    points = new String[dfds.length];
  }

// ----------------------------------------------------------------------------
// Methods
// ----------------------------------------------------------------------------
  
  public String graphMulti() {
    String d = "\\[\n\\begindc{\\commdiag}\n";
    d = vertices(d);
    d = edgesMulti(d);
    d += "\\enddc\n\\]";
    return d;
  }
  
  String vertices(String d) {
    //Get width.
    int width = 0;
    for (int i = 0; i < dag.length; i++) {
      if (dag[i].length>width) {width = dag[i].length;}
    }
    //Place vertices.
    int k; int n = width; int h = dag.length; String x,y; int z;
    for (int i = 0; i < h; i++) {
      k = dag[i].length;
      for (int j = 0; j < k; j++) {
        z = dag[i][j];
        x = Integer.toString(n-k+2*j);
        y = Integer.toString(2*h-2-2*i);
        points[z] = "A" + x + y;
        d += "\\obj(" + x + "," + y + ")[" + points[z] + "]";
        d += "{$(" + dfds[z] + ":" + arities[z] + ")$}\n";
      }
    }
    return d;
  }
  
  String edgesMulti(String d) {
    for (int i = 1; i < parentsMulti.length; i++) {
      for (int j = 0; j < parentsMulti[i].length; j++) {
        d += "\\mor{" + points[i] + "}{" +
             points[parentsMulti[i][j]] + "}";
      d += "{}[\\atright,\\solidarrow]\n";
      }
    }
    return d;
  }

// ----------------------------------------------------------------------------
}  // ----------------------------------------------------------------------
// ---------------------------------------------------------------------------- 