package expandDG;

import org.w3c.dom.*;

public class FormulaBuilder
//extends DictionaryOwner
{

// ----------------------------------------------------------------------------
// Data
// ----------------------------------------------------------------------------
  
  public MathDefDictionary mdd;
  public Document d;
  public String[] f = new String[0];
  public String[] b = new String[0];
  
  public String[] auxData;
  public int lineLength = 80;
  
  private String[] freshVarArray = {"x", "y", "z"};
  
  private final int definitionXMLplace = 0;
  private final int definiendumXMLplace = 1;
  private final int definiensXMLplace = 2;

// ----------------------------------------------------------------------------
// Constructors
// ----------------------------------------------------------------------------
  
  public FormulaBuilder(String s) {
    this.mdd = new MathDefDictionary();
    this.d = mdd.getNewDocument();
    Element root = d.createElement(s);
    d.appendChild(root);
  }
  
  public FormulaBuilder() {
    this.mdd = new MathDefDictionary();
    this.d = mdd.getNewDocument();
    Element root = d.createElement("formula");
    d.appendChild(root);
  }
  
  public FormulaBuilder(MathDefDictionary m) {
    this.mdd = m;
    this.d = mdd.getNewDocument();
    Element root = d.createElement("formula");
    d.appendChild(root);
  }

// ----------------------------------------------------------------------------
// Methods
// ----------------------------------------------------------------------------
  
  public String toString() {
    String s = "";
    NodeList n = this.d.getFirstChild().getChildNodes();
    Element e; String y,t,type,num;
    for (int i = 0; i < n.getLength(); i++) {
      e = (Element) n.item(i);
      y = e.getTagName();
      if (y.equals("text")) {
        t = e.getFirstChild().getNodeValue();
        s += t;
      }
      if (y.equals("var")) {
        type = e.getAttribute("t");
        num = e.getAttribute("n");
        t = null;
        if (type.equals("b")) {t = this.b[Integer.parseInt(num)];}
        if (type.equals("f")) {t = this.f[Integer.parseInt(num)];}
        s += t;
      }
      if (y.equals("def")) {
        t = e.getFirstChild().getNodeValue();
        s += "\\mathop{\\mathtt{" + t + "}}";
      }
    }
    return s;
  }
  
  public String toStringNaked() {
    String s = "";
    NodeList n = this.d.getFirstChild().getChildNodes();
    Element e; String y,t,type,num;
    for (int i = 0; i < n.getLength(); i++) {
      e = (Element) n.item(i);
      y = e.getTagName();
      if (y.equals("text")) {
        t = e.getFirstChild().getNodeValue();
        s += t;
      }
      if (y.equals("var")) {
        type = e.getAttribute("t");
        num = e.getAttribute("n");
        t = null;
        if (type.equals("b")) {t = this.b[Integer.parseInt(num)];}
        if (type.equals("f")) {t = this.f[Integer.parseInt(num)];}
        s += t;
      }
      if (y.equals("def")) {
        t = e.getFirstChild().getNodeValue();
        s += t;
      }
    }
    return s;
  }
  
  public String[] toStringForYAFE() {
    String[] a = null;
    NodeList n = this.d.getFirstChild().getChildNodes();
    Element e; String y,t,type,num;
    for (int i = 0; i < n.getLength(); i++) {
      e = (Element) n.item(i);
      y = e.getTagName();
      if (y.equals("text")) {
        t = e.getFirstChild().getNodeValue();
        a = concatString(a,t);
      }
      if (y.equals("var")) {
        type = e.getAttribute("t");
        num = e.getAttribute("n");
        t = null;
        if (type.equals("b")) {t = this.b[Integer.parseInt(num)];}
        if (type.equals("f")) {t = this.f[Integer.parseInt(num)];}
        a = concatString(a,t);
      }
      if (y.equals("def")) {
        t = e.getFirstChild().getNodeValue();
        a = concatString(a,t);
      }
    }
    return a;
  }
  
  public String toStringWithWrap() {
    return toStringWithWrap(this.lineLength);
  }
  
  public String toStringWithWrap(int lineLength) {
    String t = toString();
    return lineWrap(t,lineLength);
  }
  
  public String lineWrap(String t, int lineLength) {
    String s = "";
    while (t.length()>0) {
      if (t.length()<=lineLength) {s+=t; t="";}
      else {
        boolean found = false; int w = lineLength;
        while (!found && w >=0) {
          if (t.charAt(w)==' ') {
            found = true;
          } else {
            w--;
          }
        }
        if (w>=0) {
          s += t.substring(0,w) + "\n";
          t = t.substring(w+1,t.length());
        } else {
          found = false; w = lineLength;
          while (!found && w < t.length()) {
            if (t.charAt(w)==' ') {
              found = true;
            } else {
              w++;
            }
          }
          if (w<t.length()) {
            s += t.substring(0,w) + "\n";
            t = t.substring(w+1,t.length());
          } else {
            s += t; t="";
          }
        }
      }
    }
    return s;
  }
  
  public void writeInternalVars() {
    Element e;
    e = this.d.createElement("FreeVariables");
    for (int i = 0; i < f.length; i++) {
      e.appendChild(d.createTextNode(f[i] + (i<f.length-1?", ":"") ));
    }
    d.getFirstChild().appendChild(e);
    e = this.d.createElement("BoundVariables");
    for (int i = 0; i < b.length; i++) {
      e.appendChild(d.createTextNode(b[i] + (i<b.length-1?", ":"") ));
    }
    d.getFirstChild().appendChild(e);
  }
  
  public void addText(String text) {
    Element e = this.d.createElement("text");
    e.appendChild(this.d.createTextNode(text));
    this.d.getFirstChild().appendChild(e);
  }
  
  public void addVar(String var) {
    //We assume this is a free variable, and add it
    //to the f array if it is not already there.
    Element e = this.d.createElement("var");
    e.setAttribute("t","f");
    int z = findString(f,var);
    if (z<0) {
      f = concatString(f,var);
      e.setAttribute("n",Integer.toString(f.length-1));
    } else {
      e.setAttribute("n",Integer.toString(z));
    }
    this.d.getFirstChild().appendChild(e);
  }
  
  public void addVar(String type, String num) {
    //Used by DefiniensNode. Does not alter f or b arrays.
    Element e = this.d.createElement("var");
    e.setAttribute("t",type);
    e.setAttribute("n",num);
    this.d.getFirstChild().appendChild(e);
    //Diagnostic:
    System.out.println("<FormulaBuilder.java:198>Added variable of type " +
                       type + " and number " + num + ".");
    //
  }
  
  public void addDef(String def) {
    Element e = this.d.createElement("def");
    e.appendChild(this.d.createTextNode(def));
    this.d.getFirstChild().appendChild(e);
  }
  
  public void append(FormulaBuilder x) {
    x.prefixVars(this);
    this.b = new String[x.b.length];
    System.arraycopy(x.b,0,this.b,0,x.b.length);
    this.f = new String[x.f.length];
    System.arraycopy(x.f,0,this.f,0,x.f.length);
    this.suffixDoc(x);
  }

  public int bind(String v) {
    int z = -1;
    for (int i = 0; i < f.length; i++) {
      if (f[i].equals(v)) {
        z = i;
        //Delete this free variable.
        String[] a = new String[f.length - 1];
        System.arraycopy(f,0,a,0,z);
        System.arraycopy(f,z+1,a,z,f.length-(z+1));
        f = a;
        //Add it as a bound variable.
        b = concatString(b,v);
        //Update document.
        NodeList n = d.getFirstChild().getChildNodes();
        Element e; String s;
        for (int j = 0; j < n.getLength(); j++) {
          e = (Element) n.item(j);
          s = e.getTagName();
          if (s.equals("var")) {
            String type = e.getAttribute("t");
            String num = e.getAttribute("n");
            if (type.equals("f")) {
              int m = Integer.parseInt(num);
              if (m==z) {
                e.setAttribute("t","b");
                e.setAttribute("n",Integer.toString(b.length-1));
              }
              if (m>z) {
                e.setAttribute("n",Integer.toString(m-1));
              }
            }
          }
        }
      }
    }
    return z;
  }
  
  public String fresh() {
    int L = freshVarArray.length;
    int[][] subs = new int[L][];
    int[] minors = new int[L];
    for (int i = 0; i < L; i++) {
      String v = freshVarArray[i];
      int z = -1;
      for (int j = 0; j < this.b.length; j++) {
        z = freshType(this.b[j],v);
        if (z>=0) { subs[i] = concatInt(subs[i],z); }
      }
      for (int j = 0; j < this.f.length; j++) {
        z = freshType(this.f[j],v);
        if (z>=0) { subs[i] = concatInt(subs[i],z); }
      }
      minors[i] = minNonNegDiff(subs[i]);
    }
    int letter = 0;
    int index = minors[letter];
    for (int i = 1; i < L; i++) {
      if (minors[i]<index) {
        letter = i;
        index = minors[letter];
      }
    }
    String u = freshVarArray[letter];
    String w = Integer.toString(index);
    return(u + "_{" + w + "}");
  }
  
  public int freshType(String s, String v) {
    int y = -1;
    if (s.length() >= 5) {
      if (s.substring(0,3).equals(v+"_{")) {
        String n = s.substring(3,s.length()-1);
        try {
          y = Integer.parseInt(n);
        } catch (NumberFormatException nfe) {
        }
      }
    }
    return y;
  }
  
  public int minNonNegDiff(int[] a) {
    int z = 0;
    if (a!=null) {
      z = -1;
      boolean different = false;
      while (!different) {
        z++;
        boolean maybe = true;
        for (int i = 0; i < a.length; i++) {
          maybe &= (a[i]!=z);
        }
        different = maybe;
      }
    }
    return z;
  }
  
  public FormulaBuilder plugIn(FormulaBuilder P, FormulaBuilder[] t) {
    FormulaBuilder f = new FormulaBuilder();
    //Adjust variables:
    for (int i = 0; i < t.length; i++) {
      f.suffixVars(t[i]);
      for (int j = 0; j < i; j++) {
        t[i].prefixVars(t[i-1-j]);
      }
    }
    //Prepare to manage bound vars of P:
    int z = P.countBV();
    String[] BVnames = new String[z];
    boolean[] initialized = new boolean[z];
    for (int i = 0; i < z; i++) {initialized[i] = false;}
    //Process the document in P:
    NodeList n = P.d.getFirstChild().getChildNodes();
    Element e,k; String s;
    for (int i = 0; i < n.getLength(); i++) {
      e = (Element) n.item(i);
      s = e.getTagName();
      if (s.equals("text")) {
        k = (Element) f.d.createElement(s);
        String txt = e.getFirstChild().getNodeValue();
        k.appendChild(f.d.createTextNode(txt));
        f.d.getFirstChild().appendChild(k);
      }
      if (s.equals("def")) {
        k = (Element) f.d.createElement(s);
        String txt = e.getFirstChild().getNodeValue();
        k.appendChild(f.d.createTextNode(txt));
        f.d.getFirstChild().appendChild(k);
      }
      if (s.equals("var")) {
        String type = e.getAttribute("t");
        String num = e.getAttribute("n");
        int y = Integer.parseInt(num);
        if (type.equals("f")) {
          f.suffixDoc(t[y]);
        }
        if (type.equals("b")) {
          if (!initialized[y]) {
            BVnames[y] = f.fresh();
            initialized[y] = true;
          }
          f.addVar(BVnames[y]);
        }
      }
    }
    //Bind variables:
    for (int i = 0; i < BVnames.length; i++) {
      f.bind(BVnames[i]);
    }
    return f;
  }
  
  public int countBV() {
    String[] a = new String[0];
    NodeList n = this.d.getFirstChild().getChildNodes();
    Element e; String s; int z;
    for (int i = 0; i < n.getLength(); i++) {
      e = (Element) n.item(i);
      s = e.getTagName();
      if (s.equals("var")) {
        String type = e.getAttribute("t");
        if (type.equals("b")) {
          String num = e.getAttribute("n");
          z = findString(a,num);
          if (z<0) {a=concatString(a,num);}
        }
      }
    }
    return a.length;
  }
  
  public int countFV() {
    String[] a = new String[0];
    NodeList n = this.d.getFirstChild().getChildNodes();
    Element e; String s; int z;
    for (int i = 0; i < n.getLength(); i++) {
      e = (Element) n.item(i);
      s = e.getTagName();
      if (s.equals("var")) {
        String type = e.getAttribute("t");
        if (type.equals("f")) {
          String num = e.getAttribute("n");
          z = findString(a,num);
          if (z<0) {a=concatString(a,num);}
        }
      }
    }
    return a.length;
  }
  
  public void suffixVars(FormulaBuilder x) {
    //Disjoint union of bound variables:
    String[] c = new String[this.b.length + x.b.length];
    System.arraycopy(this.b,0,c,0,this.b.length);
    System.arraycopy(x.b,0,c,this.b.length,x.b.length);
    this.b = c;
    //Union of free variables:
    for (int i = 0; i < x.f.length; i++) {
      boolean present = false;
      for (int j = 0; j < this.f.length; j++) {
        if (this.f[j].equals(x.f[i])) {present = true;}
      }
      if (!present) {
        this.f = concatString(this.f, x.f[i]);
      }
    }
  }
  
  public void prefixVars(FormulaBuilder x) {
    int base = x.b.length;
    //Disjoint union of bound variables:
    String[] c = new String[x.b.length + this.b.length];
    System.arraycopy(x.b,0,c,0,x.b.length);
    System.arraycopy(this.b,0,c,x.b.length,this.b.length);
    this.b = c;
    //Union of free variables:
    String[] g = new String[x.f.length];
    System.arraycopy(x.f,0,g,0,x.f.length);
    for (int i = 0; i < this.f.length; i++) {
      boolean present = false;
      for (int j = 0; j < g.length; j++) {
        if (g[j].equals(this.f[i])) {present = true;}
      }
      if (!present) {
        g = concatString(g, this.f[i]);
      }
    }
    //Save old free variable array:
    String[] h = new String[this.f.length];
    System.arraycopy(this.f,0,h,0,this.f.length);
    //Complete the union operation:
    this.f = g;
    //Update of variable numbers:
    NodeList n = this.d.getFirstChild().getChildNodes();
    Element e; String s;
    for (int i = 0; i < n.getLength(); i++) {
      e = (Element) n.item(i);
      s = e.getTagName();
      if (s.equals("var")) {
        String type = e.getAttribute("t");
        if (type.equals("b")) {
          String oldnum = e.getAttribute("n");
          int z = base + Integer.parseInt(oldnum);
          e.setAttribute("n",Integer.toString(z));
        }
        if (type.equals("f")) {
          String oldnum = e.getAttribute("n");
          String var = h[Integer.parseInt(oldnum)];
          int z = findString(this.f,var);
          e.setAttribute("n",Integer.toString(z));
        }
      }
    }
  }
  
  public void suffixDoc(FormulaBuilder x) {
    NodeList n = x.d.getFirstChild().getChildNodes();
    Element e,f; String s;
    for (int i = 0; i < n.getLength(); i++) {
      e = (Element) n.item(i);
      s = e.getTagName();
      if (s.equals("text")) {
        f = (Element) this.d.createElement(s);
        String t = e.getFirstChild().getNodeValue();
        f.appendChild(this.d.createTextNode(t));
        this.d.getFirstChild().appendChild(f);
      }
      if (s.equals("var")) {
        f = (Element) this.d.createElement(s);
        f.setAttribute("t",e.getAttribute("t"));
        f.setAttribute("n",e.getAttribute("n"));
        this.d.getFirstChild().appendChild(f);
      }
      if (s.equals("def")) {
        f = (Element) this.d.createElement(s);
        String t = e.getFirstChild().getNodeValue();
        f.appendChild(this.d.createTextNode(t));
        this.d.getFirstChild().appendChild(f);
      }
    }
  }
  
  public void suffixDoc(Node N) {
    NodeList n = N.getChildNodes();
    Element e,f; String s;
    for (int i = 0; i < n.getLength(); i++) {
      e = (Element) n.item(i);
      s = e.getTagName();
      if (s.equals("text")) {
        f = (Element) this.d.createElement(s);
        String t = e.getFirstChild().getNodeValue();
        f.appendChild(this.d.createTextNode(t));
        this.d.getFirstChild().appendChild(f);
      }
      if (s.equals("var")) {
        f = (Element) this.d.createElement(s);
        f.setAttribute("t",e.getAttribute("t"));
        f.setAttribute("n",e.getAttribute("n"));
        this.d.getFirstChild().appendChild(f);
      }
      if (s.equals("def")) {
        f = (Element) this.d.createElement(s);
        String t = e.getFirstChild().getNodeValue();
        f.appendChild(this.d.createTextNode(t));
        this.d.getFirstChild().appendChild(f);
      }
    }
  }
  
  public Element suffixDoc(Element p, FormulaBuilder x) {
    NodeList n = x.d.getFirstChild().getChildNodes();
    Element e,f; String s;
    for (int i = 0; i < n.getLength(); i++) {
      e = (Element) n.item(i);
      s = e.getTagName();
      if (s.equals("text")) {
        f = (Element) this.d.createElement(s);
        String t = e.getFirstChild().getNodeValue();
        f.appendChild(this.d.createTextNode(t));
        p.appendChild(f);
      }
      if (s.equals("var")) {
        f = (Element) this.d.createElement(s);
        f.setAttribute("t",e.getAttribute("t"));
        f.setAttribute("n",e.getAttribute("n"));
        p.appendChild(f);
      }
      if (s.equals("def")) {
        f = (Element) this.d.createElement(s);
        String t = e.getFirstChild().getNodeValue();
        f.appendChild(this.d.createTextNode(t));
        p.appendChild(f);
      }
    }
    return p;
  }

  public int findString(String[] a, String s) {
    int z = -1;
    if (a != null) {
      for (int i = 0; i < a.length; i++) {
        if (a[i].equals(s)) { z = i; break; }
      }
    }
    return z;
  }

  public String[] concatString(String[] a, String s) {
    if (a == null) {
      a = new String[1];
      a[0] = s;
    } else {
      String[] b = new String[a.length + 1];
      System.arraycopy(a,0,b,0,a.length);
      b[a.length] = s;
      a = b;
    }
    return a;
  }
  
  public int[] concatInt(int[] a, int z) {
    if (a==null) {
      a = new int[1];
      a[0] = z;
    } else {
      int[] b = new int[a.length+1];
      System.arraycopy(a,0,b,0,a.length);
      b[a.length] = z;
      a = b;
    }
    return a;
  }

// ----------------------------------------------------------------------------
}  // ----------------------------------------------------------------------
// ---------------------------------------------------------------------------- 