package graphs;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.text.*;
import javax.swing.filechooser.*;

public class Kmap
  extends GraphViewer
  implements ActionListener, MouseListener, MouseMotionListener {

// ----------------------------------------------------------------------------
// Data
// ----------------------------------------------------------------------------
  
  JFileChooser fc;
  JScrollPane scrollPane;
  JTextField textField;
  
  JPopupMenu popup;
  JMenuItem expand, collapse, sideline, unsideline;
  
  private KmapVertex[] vertices;
  private Edge[] edges;
  
  private String dfd, arity;
  private KmapDager kmd;
  
  private boolean clear;
  private int maybeDragging = -1; //Should rename to simply "dragging".
  private int dx, dy;
  private KmapVertex ghost;
  private int lastVertexUnderPoint;
  
  private int pad = 4;
  private int sidelineTop;
  
  private final double THETA = Math.PI/6;
  private int headRadius = 10;
  private final double SINTHETA = Math.sin(THETA);
  private final double COSTHETA = Math.cos(THETA);
  
  private boolean talk = true; //controls diagnostic output

// ----------------------------------------------------------------------------
// Constructors
// ----------------------------------------------------------------------------

  public Kmap() {
    textField = new JTextField(100);
    textField.addActionListener(this);
    fc = new JFileChooser();
    scrollPane = new JScrollPane(this);
    scrollPane.addMouseListener(this);
    scrollPane.addMouseMotionListener(this);
    
    popup = new JPopupMenu();
    
    expand = new JMenuItem("Expand");
    expand.addActionListener(this);
    popup.add(expand);
    
    collapse = new JMenuItem("Collapse");
    collapse.addActionListener(this);
    popup.add(collapse);
    
    sideline = new JMenuItem("Sideline");
    sideline.addActionListener(this);
    popup.add(sideline);
    
    unsideline = new JMenuItem("Unsideline");
    unsideline.addActionListener(this);
    popup.add(unsideline);
  }

// ----------------------------------------------------------------------------
// Methods
// ----------------------------------------------------------------------------

  public void paint(Graphics g) {
    if (clear) {
      int w = this.getSize().width;
      int h = this.getSize().height;
      //Diagnostic:
      if (talk && false) System.out.println("<Kmap:64>Kmap width: " +
                                   Integer.toString(w) + ", height: " +
                                   Integer.toString(h));
      //
      g.clearRect(0,0,w,h);
      clear = false;
    }
    if (edges!=null) {
      Point q,r;
      for (int i = 0; i < edges.length; i++) {
        if (!edges[i].tail.isSidelined() &&
            !edges[i].head.isSidelined()) {
          q = edges[i].getTailPosition();
          r = edges[i].getHeadPosition();
          g.drawLine(q.x,q.y,r.x,r.y);
          Point[] head = makeHead(edges[i]);
          g.drawLine(head[0].x,head[0].y,head[1].x,head[1].y);
          g.drawLine(head[0].x,head[0].y,head[2].x,head[2].y);
        }
      }
    }
    if (vertices!=null) {
      String t; Point p; Dimension d;
      for (int i = 0; i < vertices.length; i++) {
        t = vertices[i].getText();
        if (vertices[i].isSidelined()) {
          p = new Point();
          p.x = vertices[i].sideX;
          p.y = vertices[i].sideY;
        } else {
          p = vertices[i].getPosition();
        }
        d = vertices[i].getSize();
        g.clearRect(p.x-pad,p.y-d.height-pad+2,d.width+2*pad,d.height+2*pad);
        g.drawString(t,p.x,p.y);
      }
    }
    if (maybeDragging>=0) {
      String t = ghost.getText();
      Point p = ghost.getPosition();
      g.setColor(Color.GRAY);
      g.drawString(t,p.x,p.y);
      g.setColor(Color.BLACK);
    }
  }
  
  Point[] makeHead(Edge E) {
    Point f = E.getTailPosition();
    Point e = E.getHeadPosition();
    int w = E.head.getSize().width;
    int h = E.head.getSize().height;
    float a1 = E.head.X-pad; int a2 = E.head.Y-h-pad+2;
    float b1 = E.head.X+w+pad; int b2 = E.head.Y-h-pad+2;
    int c1 = E.head.X-pad; int c2 = E.head.Y+pad+2;
    int d1 = E.head.X+w+pad; int d2 = E.head.Y+pad+2;
    float phi = ((b2-c2)/(b1-c1))*(f.x-c1)+c2;
    float psi = ((a2-d2)/(a1-d1))*(f.x-d1)+d2;
    float u,v;
    //Diagnostic:
    if (talk && false) {System.out.println("phi: " + Float.toString(phi) +
                                  "psi: " + Float.toString(psi) +
                                  "f.x: " + Integer.toString(f.x) +
                                  "\n" +
                                  Float.toString((b2-c2)/(b1-c1)) + " " +
                                  Float.toString((a2-d2)/(a1-d1)) + "\n" +
                                  Integer.toString(f.x-c1) + " " +
                                  Integer.toString(f.x-d1)
                                  );}
    //
    if (f.y <= phi && f.y < psi) {
      //Diagnostic:
      if (talk && false) {System.out.println("Region 1.");}
      //
      u = (a2-e.y)*(f.x-e.x)/(f.y-e.y) + e.x; v = a2;
    } else if (f.y <= psi && f.y > phi) {
      //Diagnostic:
      if (talk && false) {System.out.println("Region 2.");}
      //
      u = b1; v = (b1-e.x)*(f.y-e.y)/(f.x-e.x) + e.y;
    } else if (f.y >= phi && f.y > psi) {
      //Diagnostic:
      if (talk && false) {System.out.println("Region 3.");}
      //
      u = (c2-e.y)*(f.x-e.x)/(f.y-e.y) + e.x; v = c2;
    } else {
      //Diagnostic:
      if (talk && false) {System.out.println("Region 4.");}
      //
      u = a1; v = (a1-e.x)*(f.y-e.y)/(f.x-e.x) + e.y;
    }
    //Diagnostic:
    if (talk && false) {
      System.out.println("As float u is " + Float.toString(u) + ".\n" +
                         "When cast to int: " + Integer.toString((int) u) + ".");
    }
    //
    Point g = new Point((int) u,(int) v);
    double k = headRadius/distance(f,g);
    double z1 = k*(f.x-g.x);
    double z2 = k*(f.y-g.y);
    Point[] head = new Point[3];
    head[0] = g;
    double j1 = z1*COSTHETA - z2*SINTHETA;
    double j2 = z1*SINTHETA + z2*COSTHETA;
    //Diagnostic:
    if (talk && false) {
      System.out.println("k: " + Double.toString(k) +
                         "\n z1: " + Double.toString(z1) +
                         "\n z2: " + Double.toString(z2) +
                         "\n j1: " + Double.toString(j1) +
                         "\n j2: " + Double.toString(j2));
    }
    //
    head[1] = new Point((int) j1 + g.x,(int) j2 + g.y);
    j1 = z1*COSTHETA + z2*SINTHETA;
    j2 = z2*COSTHETA - z1*SINTHETA;
    head[2] = new Point((int) j1 + g.x,(int) j2 + g.y);
    return head;
  }
  
  double distance(Point a, Point b) {
    double a1 = (double) a.x; double a2 = (double) a.y;
    double b1 = (double) b.x; double b2 = (double) b.y;
    return Math.sqrt((a1-b1)*(a1-b1)+(a2-b2)*(a2-b2));
  }
  
  int vertexUnderPoint(int x, int y) {
    int v = -1;
    if (vertices!=null) {
      for (int i = 0; i < vertices.length; i++) {
        if(vertices[i].isOver(x,y)) {
          v = i; break;
        }
      }
    }
    return v;
  }
  
  public void mouseDragged(MouseEvent e) {
    int x = e.getX(); int y = e.getY();
    if (maybeDragging>=0) {
      ghost.setPosition(x-dx,y-dy);
    }
    clear = true;
    repaint();
  }
  
  public void mouseMoved(MouseEvent e) {
  }
  
  public void mouseClicked(MouseEvent e) {
  }

  public void mouseEntered(MouseEvent e) {
  }
  
  public void mouseExited(MouseEvent e) {
  }
  
  public void mousePressed(MouseEvent e) {
    maybeShowPopup(e);
    int x = e.getX(); int y = e.getY();
    //Diagnostic:
    if (talk) {
      boolean inside = false;
      if (vertices!=null) {
        for (int i = 0; i < vertices.length; i++) {
          inside |= vertices[i].isOver(x,y);
        }
      }
      if (inside) {
        System.out.println("Mouse is over a vertex.");
      } else {
        System.out.println("Mouse is not over a vertex.");
      }
      System.out.println("Position is: " +
        Integer.toString(x) + "," + Integer.toString(y)
      );
    }
    //
    int v = vertexUnderPoint(x,y);
    if (v>=0 && e.getButton()==MouseEvent.BUTTON1) {
      maybeDragging = v;
      dx = x - vertices[v].getPosition().x;
      dy = y - vertices[v].getPosition().y;
      ghost = new KmapVertex(vertices[v].getText());
      ghost.setPosition(vertices[v].getPosition());
    }
    clear = true;
    repaint();
  }
  
  public void mouseReleased(MouseEvent e) {
    maybeShowPopup(e);
    int x = e.getX(); int y = e.getY();
    if (maybeDragging>=0) {
      int v = maybeDragging;
      boolean okay = true;
      for (int i = 0; i < vertices.length; i++) {
        if (i!=v) {
          okay &= !ghost.overlaps(vertices[i]);
          //Diagnostic:
          if (talk && false) {
            if (ghost.overlaps(vertices[i])) {
              System.out.println("<Kmap.java:201>Ghost overlaps vertex " +
                                 vertices[i].dfd + ":" + vertices[i].arity);
            }
          }
          //
        }
      }
      if (okay) {
        vertices[v].setPosition(x-dx,y-dy);
      }
      maybeDragging = -1;
    }
    clear = true;
    repaint();
  }
  
  private void maybeShowPopup(MouseEvent e) {
    if (e.isPopupTrigger()) {
      int x = e.getX(); int y = e.getY();
      int v = vertexUnderPoint(x,y);
      if (v>=0) {
        lastVertexUnderPoint = v;
        KmapVertex V = vertices[v];
        expand.setEnabled(!V.isExpanded() && !V.isSidelined());
        collapse.setEnabled(V.isExpanded());
        sideline.setEnabled(!V.isSidelined());
        unsideline.setEnabled(V.isSidelined());
        popup.show(e.getComponent(),x,y);
      }
    }
  }
  
  public void actionPerformed(ActionEvent e) {
    if (e.getSource().equals(textField)) {
      int colon = parseDfd();
      if (colon==-1) {
        JOptionPane.showMessageDialog(
          this,"Input format is '[dfd]:[arity]'",
          "Error",
          JOptionPane.ERROR_MESSAGE
        );
      } else {
        int n = JOptionPane.showConfirmDialog(
          this,"Start new graph?"
        );
        if (n==JOptionPane.YES_OPTION) {
          sidelineTop = this.getSize().height - 4;
          kmd = new KmapDager(dfd,arity);
          KmapVertex v = kmd.vertices[0];
          //Diagnostic:
          if (talk) {
            Dimension d = v.getSize();
            System.out.println(Integer.toString(d.width) + "," +
                               Integer.toString(d.height));
          }
          //
          int W = this.getSize().width;
          v.setPosition((W-v.getSize().width)/2,32);
          v.justAdded = false;
          //Diagnostic:
          if (talk) {
            int x,y,w,h;
            x = v.getPosition().x; y = v.getPosition().y;
            w = v.getSize().width; h = v.getSize().height;
            System.out.println(
              "Vertex: " + v.getText() + " has corners:\n" +
              "(" + Integer.toString(x) + "," +
              Integer.toString(y-h) + ")    (" + Integer.toString(x+w) +
              "," + Integer.toString(y-h) + ")\n" +
              "(" + Integer.toString(x) + "," +
              Integer.toString(y) + ")    (" + Integer.toString(x+w) +
              "," + Integer.toString(y) + ")"
            );
          }
          //
          vertices = kmd.vertices;
          edges = kmd.edges;
          clear = true;
          repaint();
        }
      }
      
    } else if (e.getSource().equals(expand)) {
      int v = lastVertexUnderPoint;
      if (v>=0) {
        KmapVertex kv = vertices[v];
        if (!kv.isExpanded()) {
          int k = kmd.getNumber(kv);
          kmd.expand(k);
          kv.expand();
          vertices = kmd.vertices;
          edges = kmd.edges;
          clear = true;
          repaint();
        }
      }
      
    } else if (e.getSource().equals(collapse)) {
      int v = lastVertexUnderPoint;
      if (v>=0) {
        KmapVertex kv = vertices[v];
        if (kv.isExpanded()) {
          int k = kmd.getNumber(kv);
          kmd.collapse(k);
          kv.collapse();
          //Remove deleted sideline vertices
          if (vertices!=null) {
            for (int i = 0; i < vertices.length; i++) {
              if (vertices[i].isSidelined()) {
                int z = scan(kmd.vertices,vertices[i]);
                if (z<0) {
                  removeFromSideline(vertices[i]);
                }
              }
            }
          }
          //Now update our vertex and edge arrays.
          vertices = kmd.vertices;
          edges = kmd.edges;
          clear = true;
          repaint();
        }
      }
    
    } else if (e.getSource().equals(sideline)) {
      int v = lastVertexUnderPoint;
      if (v>=0) {
        KmapVertex kv = vertices[v];
        if (!kv.isSidelined()) {
          if (kv.isExpanded()) {
            int k = kmd.getNumber(kv);
            kmd.collapse(k);
            kv.collapse();
            vertices = kmd.vertices;
            edges = kmd.edges;
          }
          kv.sideline();
          kv.sideX = 4;
          kv.sideY = sidelineTop;
          sidelineTop -= (kv.getSize().height + 4);
          clear = true;
          repaint();
        }
      }
    
    } else if (e.getSource().equals(unsideline)) {
      int v = lastVertexUnderPoint;
      if (v>=0) {
        KmapVertex kv = vertices[v];
        if (kv.isSidelined()) {
          kv.unsideline();
          removeFromSideline(kv);
          clear = true;
          repaint();
        }
      }
    }
  }
  
  void removeFromSideline(KmapVertex kv) {
    int b = kv.sideY;
    int h = kv.getSize().height + 4;
    for (int i = 0; i < vertices.length; i++) {
      if (vertices[i].isSidelined()) {
        if (vertices[i].sideY < b) {
          vertices[i].sideY += h;
        }
      }
    }
    sidelineTop += h;
  }
  
  int scan(KmapVertex[] a, KmapVertex v) {
    int z = -1;
    if (a!=null) {
      for (int i = 0; i < a.length; i++) {
        if (a[i].equals(v)) {
          z = i; break;
        }
      }
    }
    return z;
  }
  
  int parseDfd() {
    String L = textField.getText();
    int w = -1;
    for (int j = 0; j < L.length(); j++) {
      if (L.substring(j,j+1).equals(":")) {
        w = j; break;
      }
    }
    if (w>=0) {
      dfd = L.substring(0,w);
      arity = L.substring(w+1,L.length());
    }
    return w;
  }
  
  void addVertex(KmapVertex v) {
    if (vertices==null) {
      //Diagnostic:
      //if (talk) System.out.println("vertices was set to null");
      //
      vertices = new KmapVertex[1];
      vertices[0] = v;
    } else {
      KmapVertex[] b = new KmapVertex[vertices.length+1];
      System.arraycopy(vertices,0,b,0,vertices.length);
      b[vertices.length] = v;
      vertices = b;
    }
  }
  
  void setUpFrame(JFrame f) {
    //f.setJMenuBar(createMenuBar());
    f.setContentPane(createContentPane());
  }
  
  JMenuBar createMenuBar() { //Not used.
    JMenuBar menuBar;
    JMenu menu, submenu;
    JMenuItem menuItem;
    JRadioButtonMenuItem rbMenuItem;
    JCheckBoxMenuItem cbMenuItem;
    
    //Create the menu bar.
    menuBar = new JMenuBar();
    
    //Build the first menu.
    menu = new JMenu("File");
    menu.setMnemonic(KeyEvent.VK_F);
    menuBar.add(menu);
    
    menuItem = new JMenuItem("Open", KeyEvent.VK_O);
    menuItem.setAccelerator(KeyStroke.getKeyStroke(
            KeyEvent.VK_O, ActionEvent.CTRL_MASK));
    menuItem.addActionListener(this);
    menu.add(menuItem);
    
    menuItem = new JMenuItem("Write", KeyEvent.VK_W);
    menuItem.setAccelerator(KeyStroke.getKeyStroke(
            KeyEvent.VK_W, ActionEvent.CTRL_MASK));
    menuItem.addActionListener(this);
    menu.add(menuItem);
    
    //Build second menu.
    menu = new JMenu("Edit");
    menu.setMnemonic(KeyEvent.VK_E);
    menuBar.add(menu);
    
    //menuItem = new JMenuItem("Copy", KeyEvent.VK_C);
    menuItem = new JMenuItem(new DefaultEditorKit.CopyAction());
    menuItem.setText("Copy");
    menuItem.setMnemonic(KeyEvent.VK_C);
    menuItem.setAccelerator(KeyStroke.getKeyStroke(
            KeyEvent.VK_C, ActionEvent.CTRL_MASK));
    //menuItem.addActionListener(this);
    menu.add(menuItem);
    
    //Build third menu.
    menu = new JMenu("Expand");
    menu.setMnemonic(KeyEvent.VK_X);
    menuBar.add(menu);
    
    menuItem = new JMenuItem("Expand All", KeyEvent.VK_E);
    menuItem.setAccelerator(KeyStroke.getKeyStroke(
            KeyEvent.VK_E, ActionEvent.CTRL_MASK));
    menuItem.addActionListener(this);
    menu.add(menuItem);
    
    menuItem = new JMenuItem("Collapse All", KeyEvent.VK_C);
    menuItem.setAccelerator(KeyStroke.getKeyStroke(
            KeyEvent.VK_Q, ActionEvent.CTRL_MASK));
    menuItem.addActionListener(this);
    menu.add(menuItem);
    
    return menuBar;
  }
  
  Container createContentPane() {
    JPanel contents = new JPanel(new GridBagLayout());
    GridBagConstraints c = new GridBagConstraints();
    c.gridwidth = GridBagConstraints.REMAINDER;
    
    c.fill = GridBagConstraints.HORIZONTAL;
    contents.add(textField, c);

    c.fill = GridBagConstraints.BOTH;
    c.weightx = 1.0;
    c.weighty = 1.0;
    contents.add(scrollPane, c);
    
    contents.setPreferredSize(new Dimension(1000,600));
    return contents;
  }

// ----------------------------------------------------------------------------
}  // ----------------------------------------------------------------------
// ---------------------------------------------------------------------------- 