package graphs;

import lpt2dzfc.MathDefDictionary;
import expand.ArgumentLocator;
import util.arrapp.Arrapp;
import org.w3c.dom.*;

public class KmapDager{

// ----------------------------------------------------------------------------
// Data
// ----------------------------------------------------------------------------
  
  Node N;
  NodeList dfn;
  KmapVertex[] vertices;
  Edge[] edges;
  MathDefDictionary mdd = new MathDefDictionary();
  
  int fixedGap = 48;
  int fixedDrop = 48;
  
  boolean talk = true; //Controls diagnostic output.

// ----------------------------------------------------------------------------
// Constructors
// ----------------------------------------------------------------------------

  public KmapDager(String d, String a) {
    getPlace(d,a);
  }

// ----------------------------------------------------------------------------
// Methods
// ----------------------------------------------------------------------------

  int getNumber(KmapVertex v) {
    int z = -1;
    for (int i = 0; i < vertices.length; i++) {
      if (vertices[i].equals(v)) {z=i; break;}
    }
    return z;
  }

  int expand(int n) {
  /** ***************************************************************
   *  The passed integer n points to the desired dfd in the vertices
   *  array. The returned integer is an error level.
   ****************************************************************** */
   N = null;
   N = mdd.lookUpDefiniens(vertices[n].dfd,vertices[n].arity);
   if (N==null) {return -1;}
   else {
     dfn = N.getChildNodes();
     int N = getChildren(n);
     //Set the positions of the new vertices.
     if (N>0) {
       int width = 0;
       for (int i = 0; i < vertices.length; i++) {
         if (vertices[i].justAdded) {
           width += vertices[i].getSize().width +
                    (i<vertices.length-1?fixedGap:0);
         }
       }
       int midpoint = vertices[n].X+vertices[n].getSize().width/2;
       int x = midpoint - width/2;
       int y = vertices[n].Y+fixedDrop;
       for (int i = 0; i < vertices.length; i++) {
         if (vertices[i].justAdded) {
           vertices[i].Y = y;
           vertices[i].X = x;
           //Diagnostic:
           if (talk) {
             System.out.println("<KmapDager.java:72>Just updated position"+
                                " of vertex number " +
                                Integer.toString(i) + ".");
           }
           //
           x += vertices[i].getSize().width + fixedGap;
           vertices[i].justAdded = false;
         }
       }
     }
     return 0;
   }
  }
  
  void collapse(int n) {
    //First disown.
    disown(n);
    boolean done = false;
    while (!done) {
      done = true;
      for (int i = 1; i < vertices.length; i++) {
        if (vertices[i].isOrphan() && vertices[i].newOrphan) {
          done = false;
          vertices[i].newOrphan = false;
          disown(i);
        }
      }
    }
    //Now clear the orphans.
    //Start by counting them (ignoring the root vertex, so start at i = 1).
    int c = 0;
    for (int i = 1; i < vertices.length; i++) {
      if (vertices[i].isOrphan()) {c++;}
    }
    //Now clear.
    int k = 0;
    KmapVertex[] a = new KmapVertex[vertices.length-c];
    a[0] = vertices[0]; //We always keep the root vertex.
    for (int i = 1; i < vertices.length; i++) {
      if (vertices[i].isOrphan()) {
        k++;
      }
      else {a[i-k]=vertices[i];}
    }
    vertices = a;
  }
  
  void disown(int n) {
    //First delete edges with vertex n as tail.
    KmapVertex v = vertices[n];
    if (v.edges!=null) {
      Edge[] E = new Edge[v.edges.length];
      System.arraycopy(v.edges,0,E,0,v.edges.length);
      for (int i = 0; i < E.length; i++) {
        if (E[i].tail.equals(v)) {
          //Diagnostic:
          if (talk) {
            System.out.println("<KmapDager.java:123>Removing edge from vertex " +
                               v.dfd + ":" + v.arity + ", edge number " +
                               Integer.toString(i) + " of " +
                               Integer.toString(E.length));
          }
          //
          Vertex w = E[i].head;
          Edge e = E[i];
          v.removeEdge(e);
          w.removeEdge(e);
          this.removeEdge(e);
        }
      }
    }
    //Now remove vertex n as a parent from any and all vertices.
    for (int i = 0; i < vertices.length; i++) {
      vertices[i].parents = subtract(vertices[i].parents,n);
    }
  }
    
  void removeEdge(Edge e) {
  /** This method assumes there are no repeats in the edges array. */
    if (edges!=null) {
      boolean present = false;
      for (int i = 0; i < edges.length; i++) {
        present |= (edges[i].equals(e));
      }
      if (present) {
        Edge[] f = new Edge[edges.length-1];
        int passed = 0;
        for (int i = 0; i < edges.length; i++) {
          if (edges[i].equals(e)) {
            passed = 1;
          } else {
            f[i-passed] = edges[i];
          }
        }
        edges = f;
      }
    }
  }
  
  int getChildren(int n) {
  //Returns the number of children added.
    int N = 0;
    Element e,d; String tag;
    for (int i = 0; i < dfn.getLength(); i++) {
      e = (Element) dfn.item(i);
      tag = e.getTagName();
      if (tag.equals("def")) {
        int[] endpts = new int[0];
        if (i<dfn.getLength()-1) {
          d = (Element) dfn.item(i+1);
          tag = d.getTagName();
          if (tag.equals("text")) {
            String u = d.getFirstChild().getNodeValue();
            if (u.equals("(") || u.equals("[")) {
              ArgumentLocator argloc = new ArgumentLocator(dfn,i+2);
              endpts = argloc.locate();
            }
          }
        }
        String s = e.getFirstChild().getNodeValue();
        String a = Integer.toString(endpts.length/2);
        int childPlace = getPlace(s,a);
        N++;
        vertices[childPlace].parents =
          Arrapp.union(vertices[childPlace].parents,n);
        //Add edge
        Edge ed = new Edge(vertices[n], vertices[childPlace]);
        vertices[n].addEdge(ed);
        vertices[childPlace].addEdge(ed);
        this.addEdge(ed);
      }
    }
    return N;
  }
  
  void addEdge(Edge e) {
    if (edges==null) {
      edges = new Edge[1];
      edges[0] = e;
    } else {
      Edge[] a = new Edge[edges.length+1];
      System.arraycopy(edges,0,a,0,edges.length);
      a[edges.length] = e;
      edges = a;
    }
  }
  
  int getPlace(String d, String a) {
  /** ***************************************************************
   *  Searches for (d,a) in (dfds,arities). If not found, it is added.
   *  In any case, this method returns the index at which (d,a) can
   *  (now) be found. Also updates the parents array as necessary
   *  to keep it the same length as the dfds and arities arrays.
   ****************************************************************** */
    int z = -1;
    if (vertices!=null) {
      for (int i = 0; i < vertices.length; i++) {
        if (vertices[i].dfd.equals(d) && vertices[i].arity.equals(a)) {
          z = i; break;
        }
      }
    }
    if (z<0) {
      KmapVertex v = new KmapVertex(d+":"+a);
      v.justAdded = true;
      v.dfd = d; v.arity = a;
      append(v);
      z = vertices.length - 1;
    }
    return z;
  }
  
  void append(KmapVertex v) {
    if (vertices==null) {
      vertices = new KmapVertex[1];
      vertices[0] = v;
    } else {
      KmapVertex[] b = new KmapVertex[vertices.length+1];
      System.arraycopy(vertices,0,b,0,vertices.length);
      b[vertices.length] = v;
      vertices = b;
    }
  }
  
  int[] subtract(int[] a, int z) {
    if (a!=null) {
      int c = 0;
      for (int i = 0; i < a.length; i++) {
        if (a[i]==z) c++;
      }
      int[] b = new int[a.length - c];
      c = 0;
      for (int i = 0; i < a.length; i++) {
        if (a[i]==z) {c++;}
        else {
          b[i-c] = a[i];
        }
      }
      a = b;
    }
    return a;
  }

// ----------------------------------------------------------------------------
}  // ----------------------------------------------------------------------
// ---------------------------------------------------------------------------- 