package lpt2dzfc;

public class Quant{

// ----------------------------------------------------------------------------
// Data
// ----------------------------------------------------------------------------

  public FormulaBuilder[] boundterms;
  public FormulaBuilder suchthat = null;
  public FormulaBuilder formula;
  public Infixrel ifcRel = null;
  public FormulaBuilder[] ifcTerms = null;
  public String[] fixed = new String[0];
  
  private String type;
  private String q;
  private String c;

// ----------------------------------------------------------------------------
// Constructors
// ----------------------------------------------------------------------------

  public Quant(String type) {
    this.type = type;
  }

// ----------------------------------------------------------------------------
// Methods
// ----------------------------------------------------------------------------

  public FormulaBuilder output() {
    if (type.equals("forall")) {
      this.q = "\\forall ";
      this.c = " \\rightarrow ";
      return simple();
    }
    if (type.equals("exists")) {
      this.q = "\\exists ";
      this.c = " \\wedge ";
      return simple();
    }
    if (type.equals("existsUnique")) {
      return compound();
    }
    else {return null;}
  }
  
  private FormulaBuilder simple() {
    //Determine bound variables:
    FormulaBuilder g = new FormulaBuilder();
    for (int i = 0; i < boundterms.length; i++) {g.append(boundterms[i]);}
    for (int i = 0; i < fixed.length; i++) {g.bind(fixed[i]);}
    //Prepare any infix conditions:
    FormulaBuilder ifc = null;
    if (ifcRel!=null) {
      ifc = new FormulaBuilder();
      for (int i = 0; i < boundterms.length; i++) {
        for (int j = 0; j < ifcTerms.length; j++) {
          ifc.append(ifcRel.output(boundterms[i],ifcTerms[j]));
          if (i<boundterms.length-1 || j < ifcTerms.length-1) {ifc.addText(" \\wedge ");}
        }
      }
    }
    //Build the formula:
    FormulaBuilder out = new FormulaBuilder();
    out.addText("(");
    out.addText(q);
    for (int i = 0; i < g.f.length; i++) {
      out.addVar(g.f[i]);
      if (i!=g.f.length-1) {out.addText(", ");}
    }
    out.addText(")");
    if (ifcRel==null && suchthat==null) {
      out.append(formula);
    } else {
      out.addText("(");
      if (ifcRel!=null) {out.append(ifc);}
      if (suchthat!=null) {out.append(suchthat);}
      out.addText(c);
      if (fixed.length>0) {out.addText("(");}
      out.append(formula);
      if (fixed.length>0) {out.addText(")");}
      out.addText(")");
    }
    //Bind variables:
    for (int i = 0; i < g.f.length; i++) {out.bind(g.f[i]);}
    return out;
  }
  
  private FormulaBuilder compound() {
    //Determine bound variables:
    FormulaBuilder g = new FormulaBuilder();
    for (int i = 0; i < boundterms.length; i++) {g.append(boundterms[i]);}
    for (int i = 0; i < fixed.length; i++) {g.bind(fixed[i]);}
    //Prepare any infix conditions:
    FormulaBuilder ifc = null;
    if (ifcRel!=null) {
      ifc = new FormulaBuilder();
      for (int i = 0; i < boundterms.length; i++) {
        for (int j = 0; j < ifcTerms.length; j++) {
          ifc.append(ifcRel.output(boundterms[i],ifcTerms[j]));
          if (i<boundterms.length-1 || j < ifcTerms.length-1) {ifc.addText(" \\wedge ");}
        }
      }
    }
    //Generate new bound variables:
    FormulaBuilder h = new FormulaBuilder();
    for (int i = 0; i < boundterms.length; i++) {h.append(boundterms[i]);}
    h.append(formula);
    if (ifcRel!=null) {h.append(ifc);}
    if (suchthat!=null) {h.append(suchthat);}
    String[] y = new String[g.f.length];
    for (int i = 0; i < g.f.length; i++) {
      y[i] = h.fresh();
      h.addVar(y[i]);
    }
    //Build the formula:
    FormulaBuilder out = new FormulaBuilder();
    out.addText("(\\exists ");
    for (int i = 0; i < y.length; i++) {
      out.addVar(y[i]);
      if (i!=y.length-1) {out.addText(", ");}
    }
    out.addText(")(\\forall ");
    for (int i = 0; i < g.f.length; i++) {
      out.addVar(g.f[i]);
      if (i!=g.f.length-1) {out.addText(", ");}
    }
    out.addText(")(");
    if (ifcRel!=null) {out.append(ifc); out.addText(" \\wedge ");}
    if (suchthat!=null) {out.append(suchthat); out.addText(" \\wedge ");}
    if (fixed.length>0) {out.addText("(");}
    out.append(formula);
    if (fixed.length>0) {out.addText(")");}
    out.addText(" \\rightleftarrow ");
    for (int i = 0; i < g.f.length; i++) {
      out.addVar(g.f[i]); out.addText(" = "); out.addVar(y[i]);
      if (i!=g.f.length-1) {out.addText(" \\wedge ");}
    }
    out.addText(")");
    //Bind variables:
    for (int i = 0; i < g.f.length; i++) {out.bind(g.f[i]);}
    for (int i = 0; i < y.length; i++) {out.bind(y[i]);}
    return out;
  }

// ----------------------------------------------------------------------------
}  // ----------------------------------------------------------------------
// ---------------------------------------------------------------------------- 