package util.scanner;

public class Scanner
  extends ScanningTool {

/** *******************************************************
 *  This is a utility class to be used to scan a single string.
 *  You pass that string to the constructor.
 *  Then use this class to scan the string.
 *  Then discard.
 ********************************************************** */

// ----------------------------------------------------------------------------
// Data
// ----------------------------------------------------------------------------
  
  private String s;
  public int pointer = 0;

// ----------------------------------------------------------------------------
// Constructors
// ----------------------------------------------------------------------------

  public Scanner(String s) {
    this.s = s;
  }

// ----------------------------------------------------------------------------
// Methods
// ----------------------------------------------------------------------------

  public String[] scan() {
  /** *****************************************************
   *  Call repeatedly in order to step through the string s,
   *  each time reading up to the next flag or end of the string.
   *  Returns a string array r of length 3. If any scanning
   *  takes place then the contents will be in r[0]. If a
   *  flag character is encountered then this will be in r[1].
   *  If the end of the string is reached then r[2] will say
   *  "END". Note that r[1] will be empty (not null) if there
   *  was no flag.
   ******************************************************** */
    String[] r = new String[3];
    r[0] = ""; r[1] = ""; r[2] = "";
    if (pointer >= s.length()) {
      r[2] = "END";
      return r;
    } else {
      boolean flagged = false;
      String t;
      while (!flagged) {
        t = s.substring(pointer,pointer+1);
        pointer++;
        r[0] += t;
        if (isFlag(t)) {
          flagged = true;
          r[1] = t;
        }
        if (pointer >= s.length()) {
          flagged = true;
          r[2] = "END";
        }
      }
      return r;
    }
  }
  
  public boolean isFlag(String t) {
  /** *****************************************************
   *  Determines whether the passed string is among those in
   *  the flags array stored in the superclass "ScanningTool".
   ******************************************************** */
    boolean found = false;
    for (int i = 0; i < super.flags.length; i++) {
      if (t.equals(super.flags[i])) {found = true; break;}
    }
    return found;
  }

// ----------------------------------------------------------------------------
}  // ----------------------------------------------------------------------
// ---------------------------------------------------------------------------- 